/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libmcve_main.h"
#include "mcve.h"


// Standard includes ... all platforms
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>

//Win32 includes
#ifdef WIN32
#include <io.h>
#include <winsock2.h>
#include <sys/timeb.h>
#else // !Win32
// UNIX/POSIX includes
#include <unistd.h>
#include <sys/socket.h>
#ifndef DISABLE_IP
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <sys/time.h>
#include <dirent.h>
#ifdef MCVE_PTHREAD
#include <pthread.h>
#endif // MCVE_PTHREAD

#endif //Win32

//SSL includes
#ifdef ENABLE_SSL
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#endif

#include "libmcve_utils.h"
#include "libmcve_misc.h"
#include "libmcve_parsing.h"
#include "libmcve_cdparsing.h"
#include "libmcve_locking.h"
#include "libmcve_sockets.h"
#include "libmcve_file.h"

#ifdef ENABLE_SSL
SSL_CTX *client_ctx;
#endif


/* BASE ROUTINES */

int MCVE_InitEngine(char *location)
{
  char egd[255];
  egd[0]=0;
#ifndef DISABLE_IP
  if (!MCVE_InitSockets()) return(0);
#endif
#ifdef ENABLE_SSL
  SSL_load_error_strings();
  OpenSSL_add_ssl_algorithms();
#ifdef NEED_SEED
  strcpy(egd, "/var/run/egd-pool");
  if (RAND_egd(egd) == -1)
    return(0);
#endif
  client_ctx=SSL_CTX_new(SSLv23_client_method());
  if (!SSL_CTX_load_verify_locations(client_ctx, location, NULL)) ;
  if (client_ctx == NULL) return(0);
#endif
 return(1);
}

void MCVE_DestroyEngine()
{
#ifdef ENABLE_SSL
  SSL_CTX_free(client_ctx);
#endif
}

//Initializes the MCVE Connection structure
void MCVE_InitConn(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn;
  myconn[0]=(MCVE_CONN)malloc(sizeof(_MCVE_CONN)*1);
  conn=(_MCVE_CONN *)myconn[0];

  conn->conn_method=-1;
  conn->port=-1;
  conn->ptr=-1;
  conn->timeout=-1;
  conn->blocking=0;
  conn->status=0;
  conn->verifyconn=1;
  conn->verifyssl=0;
  conn->max_conn_time=5;
  conn->inbuf=NULL;
  conn->outbuf=NULL;
  conn->error_text=NULL;
#ifdef ENABLE_SSL
  conn->ssl=NULL;
  conn->server_cert=NULL;
#endif
#ifdef MCVE_THREADSAFE
  MCVE_init_locks(myconn);
#endif
  conn->outstanding_auths=0;
  conn->queue_length=0;
  conn->queue=NULL;
  memset(conn->location, 0, 250);
}

int MCVE_SetBlocking(MCVE_CONN *myconn, int tf)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  conn->blocking=tf;
  return(1);
}

int MCVE_SetTimeout(MCVE_CONN *myconn, long timeout)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  conn->timeout=timeout;
  return(1);
}

// Sets method and path for drop files
// Returns 0 on error
int MCVE_SetDropFile(MCVE_CONN *myconn, char *df_location)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  if (strlen(df_location) > 249) return(0);

  if (!MCVE_DirectoryExists(df_location))
    return(0);

  strncpy(conn->location, df_location,250);
  conn->conn_method=MCVE_FILE;
  conn->status=MCVE_CONNECTED;
  return(1);
}



// Sets method,location, and portnum for IP
// Returns 0 on error
int MCVE_SetIP(MCVE_CONN *myconn, char *host, unsigned long port)
{
#ifndef DISABLE_IP  
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  if (strlen(host) > 249 || strlen(host) < 1) return(0);
  strncpy(conn->location, host,250);
  if (port < 1) return(0);
  conn->port=port;
  conn->conn_method=MCVE_SOCKETS;
  return(1);
#else
  return(0);
#endif


}



// Sets method and path for drop files
// Returns 0 on error
int MCVE_SetSSL(MCVE_CONN *myconn, char *host, unsigned long port)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
#ifdef ENABLE_SSL
  if (host == NULL || strlen(host) > 249 || strlen(host) < 1) {
    MCVE_Set_Conn_Error(myconn, "Invalid Host Name");  // Set Error Text
    return(0);
  }
  strncpy(conn->location, host, 250);
  if (port < 1) {
    MCVE_Set_Conn_Error(myconn, "Invalid Port Number");  // Set Error Text
    return(0);
  }
  conn->port=port;
  conn->conn_method=MCVE_SSL;
  return(1);
#else
  return(0);
#endif
}


void MCVE_VerifyConnection(MCVE_CONN *myconn, int tf)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  conn->verifyconn=tf;
}

void MCVE_VerifySSLCert(MCVE_CONN *myconn, int tf)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  conn->verifyssl=tf;
}

int MCVE_Connect(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
#ifdef ENABLE_SSL
  long err;
#endif
  if (conn->conn_method == MCVE_FILE) {
    return(1);  // No connection necessary for Drop File
#ifndef DISABLE_IP
  } else if (conn->conn_method == MCVE_SOCKETS) {
    conn->ptr=MCVE_ip_connect(myconn);
    if (conn->ptr == -1) {
       return(0);  //Connection Failed
    } 
#ifdef ENABLE_SSL
  } else if (conn->conn_method == MCVE_SSL) {
    conn->ptr=MCVE_ip_connect(myconn);
     if (conn->ptr == -1) {
        return(0);  //Connection Failed
     }
     conn->ssl=SSL_new(client_ctx);
     SSL_set_fd(conn->ssl, conn->ptr);
     if (SSL_connect(conn->ssl) == -1) {
        MCVE_Set_Conn_Error(myconn, "SSL Negotiation Failed");  // Set Error Text
	SSL_free(conn->ssl);
	if (conn->ptr != -1)
          MCVE_CloseSocket(conn->ptr);
        conn->ptr=-1;
        return(0);
     }
     conn->server_cert = SSL_get_peer_certificate(conn->ssl);
     //printf("Server Certificate subject: %s\r\n", X509_NAME_oneline(X509_get_subject_name((X509 *)conn->server_cert),0,0));
    // printf("Server Certificate issuer: %s\r\n", X509_NAME_oneline(X509_get_issuer_name((X509 *)conn->server_cert),0,0));

     if (conn->verifyssl && (err=SSL_get_verify_result(conn->ssl)) != X509_V_OK) {  // SSL Certificate check failed
       MCVE_Set_Conn_Error(myconn, "SSL Certificate verification failed");
       if (conn->ptr != -1)
         MCVE_CloseSocket(conn->ptr);
       conn->ptr=-1;
       SSL_free((SSL *)conn->ssl);
       return(0);
     }
#endif /* ENABLE_SSL */
#endif /* DISABLE_IP */
  }
  conn->status = MCVE_CONNECTED;
#ifndef DISABLE_IP
  if ((conn->conn_method == MCVE_SOCKETS || conn->conn_method == MCVE_SSL) && conn->verifyconn) {
    if (!MCVE_VerifyPing(myconn)) {
      MCVE_Set_Conn_Error(myconn, "MCVE did not respond to a PING request. Ensure proper port number and MCVE v2.1 or greater.");
      if (conn->ptr != -1)
         MCVE_CloseSocket(conn->ptr);
       conn->ptr=-1;
#ifdef ENABLE_SSL
       if (conn->conn_method == MCVE_SSL)
         SSL_free(conn->ssl);
#endif
      conn->status=MCVE_DISCONNECTED;
      return(0);
     }
   }
#endif /* DISABLE_IP */
  conn->status=MCVE_CONNECTED;
  return(1);
}

void MCVE_MaxConnTimeout(MCVE_CONN *myconn, int maxtime)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  conn->max_conn_time=maxtime;
}

char *MCVE_ConnectionError(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->error_text);
}


// Closes MCVE Connection -- and is ready for MCVE_CONN to be used again
void MCVE_DestroyConn(MCVE_CONN *myconn)
{
  long i;
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];

  for (i=0; i<conn->queue_length; i++)
      MCVE_DeleteTrans(myconn, i);

#ifdef ENABLE_SSL
  if (conn->conn_method == MCVE_SSL)
    SSL_free((SSL *)conn->ssl);
#endif
#ifdef MCVE_THREADSAFE
  MCVE_destroy_locks(myconn);
#endif
  conn->conn_method=-1;
  conn->port=-1;
  if (conn->ptr != -1)
#ifndef DISABLE_IP
    MCVE_CloseSocket(conn->ptr);
#endif
  conn->ptr=-1;
  if (conn->inbuf != NULL)
    free(conn->inbuf);
  if (conn->outbuf != NULL)
    free(conn->outbuf);
  if (conn->error_text != NULL)
    free(conn->error_text);
  conn->error_text=NULL;
  conn->inbuf=NULL;
  conn->outbuf=NULL;
  conn->timeout=0;
  conn->status=0;
  conn->blocking=0;
  conn->verifyssl=0;
  conn->verifyconn=0;
  conn->max_conn_time=5;
  conn->outstanding_auths=0;
  conn->queue_length=0;
  if (conn->queue != NULL)
    free(conn->queue);
  conn->queue=NULL;

  memset(conn->location, 0, 250);
}



int MCVE_Monitor(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int ret=0;
  
  if (conn->status == MCVE_DISCONNECTED) {
    MCVE_Set_Conn_Error(myconn, "Not Connected to MCVE");  // Set Error Text
    return(0);
  }

  if (conn->conn_method == MCVE_FILE) {
    ret=MCVE_Monitor_File(myconn);
  } else if (conn->conn_method == MCVE_SOCKETS || conn->conn_method==MCVE_SSL) {
#ifndef DISABLE_IP
    ret=MCVE_Monitor_IP(myconn);

    if (!ret) {
      close(conn->ptr);
      conn->ptr=-1;
#ifdef ENABLE_SSL
      if (conn->conn_method == MCVE_SSL) SSL_free((SSL *)conn->ssl);
#endif /*ENABLE_SSL */
      MCVE_Set_Conn_Error(myconn, "Unexpected disconnection from MCVE");  // Set Error Text
      conn->status=MCVE_DISCONNECTED;
    }
    MCVE_ProcessBuffer(myconn);
#else 
    return(0);
#endif /* DISABLE_IP */
  }
 return(ret);
}

int MCVE_TransactionsSent(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int ret=0;
  if (conn->conn_method == MCVE_FILE) {
    ret=1;
  } else {
    MCVE_lock(myconn, MCVE_LOCK_OUT);
    if (conn->outbuf == NULL) ret=1;
    MCVE_unlock(myconn, MCVE_LOCK_OUT);
  }
  return(ret);
}



void MCVE_DeleteTrans(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int i;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  if (conn->queue[identifier].transaction != NULL) {
    for (i=0; i<conn->queue[identifier].transaction_fields; i++) {
      free(conn->queue[identifier].transaction[i].key);
      free(conn->queue[identifier].transaction[i].value);
    }
    free(conn->queue[identifier].transaction);
    conn->queue[identifier].transaction=NULL;
  }
  conn->queue[identifier].transaction_fields=0;
  if (conn->queue[identifier].parsed_resp != NULL) {
    for (i=0; i<conn->queue[identifier].resp_fields; i++) {
      free(conn->queue[identifier].parsed_resp[i].key);
      free(conn->queue[identifier].parsed_resp[i].value);
    }
    free(conn->queue[identifier].parsed_resp);
    conn->queue[identifier].parsed_resp=NULL;
  }
  conn->queue[identifier].resp_fields=0;

  memset(conn->queue[identifier].identifier, 0, 29);
  conn->queue[identifier].status=MCVE_UNUSED;
  if (conn->queue[identifier].auth != NULL)
    free(conn->queue[identifier].auth);
  conn->queue[identifier].auth=NULL;
  if (conn->queue[identifier].text != NULL)
    free(conn->queue[identifier].text);
  conn->queue[identifier].text=NULL;
  if (conn->queue[identifier].item != NULL)
    free(conn->queue[identifier].item);
  conn->queue[identifier].item=NULL;
  if (conn->queue[identifier].batch != NULL)
    free(conn->queue[identifier].batch);
  conn->queue[identifier].batch=NULL;
  conn->queue[identifier].code=-1;
  conn->queue[identifier].avs=-1;
  conn->queue[identifier].cv=-1;
  conn->queue[identifier].tid=-1;
  if (conn->queue[identifier].response != NULL)
    free(conn->queue[identifier].response);
  conn->queue[identifier].response=NULL;
 // conn->queue[identifier].iscommadelimiated=0;
  if (conn->queue[identifier].separated != NULL)
    free(conn->queue[identifier].separated);
  conn->queue[identifier].separated=NULL;
  conn->queue[identifier].columns=0;
  conn->queue[identifier].rows=0;
  conn->outstanding_auths--;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
}


/* Get a new Queue Number, return the location */
long MCVE_TransNew(MCVE_CONN *myconn)
{
  long pos;
  char timeout[20];
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];

  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  pos=MCVE_OpenQueueSlot(myconn);
  conn->queue[pos].status=MCVE_NEW;
  MCVE_ClearQueueSlot(myconn, pos);
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  
  if (conn->timeout > 0) {
    sprintf(timeout, "%ld", conn->timeout);
    MCVE_TransParam_Add(myconn, pos, "timeout", timeout);
  }
  return(pos);
}

int MCVE_VerifyTrans(MCVE_CONN *myconn, long identifier)
{
  return(1);
}

int MCVE_TransSend(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int ret;

  if (!MCVE_VerifyTrans(myconn, identifier)) {
    return(0);
  }
 
  ret=MCVE_SendTransaction(myconn, identifier);

  if (ret) {
    conn->outstanding_auths++;

    /* Blocking mode is set, wait until trans is complete */
    if (conn->blocking) {
      while(MCVE_CheckStatus(myconn, identifier) == MCVE_PENDING) {
        if (!MCVE_Monitor(myconn)) {
	  return(0);
	}
	MCVE_uwait(50000);
      }
    }
  }
  return(ret);
}


char *MCVE_ResponseParam(MCVE_CONN *myconn, long identifier, char *key)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int i;

  for (i=0; i<conn->queue[identifier].resp_fields; i++) {
    if (strcasecmp(key, conn->queue[identifier].parsed_resp[i].key) == 0)
      return(conn->queue[identifier].parsed_resp[i].value);
  }
  return(NULL);
}

/* Add a custom Transaction Parameter (modules, etc) */
int MCVE_TransParam_Add(MCVE_CONN *myconn, long identifier, char *key, char *value)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int field;

  field=conn->queue[identifier].transaction_fields;

  conn->queue[identifier].transaction=(MCVE_TRANS *)realloc(conn->queue[identifier].transaction, (field+1)*sizeof(MCVE_TRANS));
  conn->queue[identifier].transaction[field].key=MC_SAFE_strdup(key);
  conn->queue[identifier].transaction[field].value=MC_SAFE_strdup(value);
  conn->queue[identifier].transaction_fields++;

  return(1);
}

char *MCVE_GetTypeString(int trans_type)
{
  char *type=NULL;

  switch(trans_type) {
    case MC_TRAN_SALE:
      type="SALE";
    break;
    case MC_TRAN_PREAUTH:
      type="PREAUTH";
    break;
    case MC_TRAN_VOID:
      type="VOID";
    break;
    case MC_TRAN_PREAUTHCOMPLETE:
      type="FORCE";
    break;
    case MC_TRAN_FORCE:
      type="FORCE";
    break;
    case MC_TRAN_OVERRIDE:
      type="OVERRIDE";
    break;
    case MC_TRAN_RETURN:
      type="RETURN";
    break;
    case MC_TRAN_SETTLE:
      type="SETTLE";
    break;
    case MC_TRAN_ADMIN:
      type="ADMIN";
    break;
    case MC_TRAN_CHKPWD:
      type="CHKPWD";
    break;
    case MC_TRAN_INCREMENTAL:
      type = "INCREMENTAL";
    break;
    case MC_TRAN_REVERSAL:
      type = "REVERSAL";
    break;
  };
  return(type);
}

char *MCVE_GetAdminTypeString(int trans_type)
{
  char *type=NULL;

  switch(trans_type) {
    case MC_ADMIN_GUT:
      type="GUT";
    break;
    case MC_ADMIN_GL:
      type="GL";
    break;
    case MC_ADMIN_GFT:
      type="GFT";
    break;
    case MC_ADMIN_QC:
      type="QC";
    break;
    case MC_ADMIN_RS:
      type="RS";
    break;
    case MC_ADMIN_BT:
      type="BT";
    break;
    case MC_ADMIN_CTH:
      type="CTH";
    break;
    case MC_ADMIN_CFH:
      type="CFH";
    break;
    case MC_ADMIN_FORCESETTLE:
      type="FORCESETTLE";
    break;
    case MC_ADMIN_SETBATCHNUM:
      type="SETBATCHNUM";
    break;
    case MC_ADMIN_RENUMBERBATCH:
      type="RENUMBERBATCH";
    break;
    case MC_ADMIN_FIELDEDIT:
      type="FIELDEDIT";
    break;
  };
  return(type);
}

char *MCVE_GetEngineAdminTypeString(int trans_type)
{
  char *type=NULL;
  switch (trans_type) {
    case MC_TRAN_CHNGPWD:
      type="CHNGPWD";
    break;
    case MC_TRAN_LISTSTATS:
      type="LISTSTATS";
    break;
    case MC_TRAN_LISTUSERS:
      type="LISTUSERS";
    break;
    case MC_TRAN_GETUSERINFO:
      type="GETUSERINFO";
    break;
    case MC_TRAN_ADDUSER:
      type="ADDUSER";
    break;
    case MC_TRAN_EDITUSER:
      type="EDITUSER";
    break;
    case MC_TRAN_DELUSER:
      type="DELUSER";
    break;
    case MC_TRAN_ENABLEUSER:
      type="ENABLEUSER";
    break;
    case MC_TRAN_DISABLEUSER:
      type="DISABLEUSER";
    break;
    case MC_TRAN_IMPORT:
      type="IMPORT";
    break;
    case MC_TRAN_EXPORT:
      type="EXPORT";
    break;
  };
  return(type);
}

char *MCVE_GetUserArgString(int arg)
{
  char *str=NULL;

  switch(arg) {
    case MC_USER_PROC:
      str="PROC";
    break;
    case MC_USER_USER:
      str="USER";
    break;
    case MC_USER_PWD:
      str="PWD";
    break;
    case MC_USER_INDCODE:
      str="INDCODE";
    break;
    case MC_USER_MERCHID:
      str="MERCHID";
    break;
    case MC_USER_BANKID:
      str="BANKID";
    break;
    case MC_USER_TERMID:
      str="TERMID";
    break;
    case MC_USER_CLIENTNUM:
      str="CLIENTNUM";
    break;
    case MC_USER_STOREID:
      str="STOREID";
    break;
    case MC_USER_AGENTID:
      str="AGENTID";
    break;
    case MC_USER_CHAINID:
      str="CHAINID";
    break;
    case MC_USER_ZIPCODE:
      str="ZIPCODE";
    break;
    case MC_USER_TIMEZONE:
      str="TIMEZONE";
    break;
    case MC_USER_MERCHCAT:
      str="MERCHCAT";
    break;
    case MC_USER_MERNAME:
      str="MERNAME";
    break;
    case MC_USER_MERCHLOC:
      str="MERCHLOC";
    break;
    case MC_USER_STATECODE:
      str="STATECODE";
    break;
    case MC_USER_PHONE:
      str="SERVICEPHONE";
    break;
    case MC_USER_SUB:
      str="SUB";
    break;
    case MC_USER_CARDTYPES:
      str="CARDTYPES";
    break;
    case MC_USER_MODE:
      str="MODE";
    break;
    case MC_USER_VNUMBER:
      str="VNUMBER";
    break;
    case MC_USER_ROUTINGID:
      str="ROUTINGID";
    break;
  };
  return(str);
}


char *MCVE_GetExchargesArgString(int excharges_type)
{
  char *str=NULL;

  switch(excharges_type) {
    case MC_EXCHARGES_REST:
      str="REST";
    break;
    case MC_EXCHARGES_GIFT:
      str="GIFT";
    break;
    case MC_EXCHARGES_MINI:
      str="MINI";
    break;
    case MC_EXCHARGES_TELE:
      str="TELE";
    break;
    case MC_EXCHARGES_OTHER:
      str="OTHER";
    break;
    case MC_EXCHARGES_LAUND:
      str="LAUND";
    break;
    case MC_EXCHARGES_NONE:
      str="NONE";
    break;
    case MC_EXCHARGES_GAS:
      str="GAS";
    break;
    case MC_EXCHARGES_MILE:
      str="MILE";
    break;
    case MC_EXCHARGES_LATE:
      str="LATE";
    break;
    case MC_EXCHARGES_1WAY:
      str="1WAY";
    break;
    case MC_EXCHARGES_VIOL:
      str="VIOL";
    break;
  }
  
  return (str);
}

char *MCVE_GetPriorityString(int priority)
{
  char *str=NULL;

  switch(priority) {
    case MC_PRIO_HIGH:
      str="HIGH";
    break;
    case MC_PRIO_NORMAL:
      str="NORMAL";
    break;
    case MC_PRIO_LOW:
      str="LOW";
    break;
  }
  
  return (str);
}

char *MCVE_GetCardTypesString(int types)
{
  char *ret=NULL;
  ret=(char *)malloc(100);
  ret[0]=0;
  
  if (types & ~MC_CARD_ALL) {
    strcat(ret, "ALL");
    return(ret);
  }
  
  if (types & ~MC_CARD_VISA) {
    strcat(ret, "+VISA");
  }
  if (types & ~MC_CARD_MC) {
    strcat(ret, "+MC");
  }
  if (types & ~MC_CARD_AMEX) {
    strcat(ret, "+AMEX");
  }
  if (types & ~MC_CARD_DISC) {
    strcat(ret, "+DISC");
  }
  if (types & ~MC_CARD_DC) {
    strcat(ret, "+DC");
  }
  if (types & ~MC_CARD_JCB) {
    strcat(ret, "+JCB");
  }
  if (types & ~MC_CARD_CB) {
    strcat(ret, "+CB");
  }
  if (types & ~MC_CARD_GIFT) {
    strcat(ret, "+GIFT");
  }
  if (types & ~MC_CARD_OTHER) {
    strcat(ret, "+OTHER");
  }
  return(ret);
}

char *MCVE_GetModeString(int types)
{
  char *ret=NULL;
  ret=(char *)malloc(100);
  ret[0]=0;
  
  if (types & ~MC_MODE_BOTH) {
    strcat(ret, "BOTH");
    return(ret);
  }
  
  if (types & ~MC_MODE_AUTH) {
    strcat(ret, "+AUTH");
  }
  if (types & ~MC_MODE_SETTLE) {
    strcat(ret, "+SETTLE");
  }
 
  return(ret);
}

/* Add a new Transaction Parameter */
int MCVE_TransParam(MCVE_CONN *myconn, long identifier, int key, ...)
{
  char key_string[100];
  char value_string[1024];
  int trans_type;
  int excharges_type;
  int temp_tf;
  char *temp_str=NULL;
  va_list ap;
  char *temp=NULL;
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];

  key_string[0]=0;
  value_string[0]=0;
  va_start(ap, key);

  switch(key) {
    case MC_TRANTYPE:
      MC_SAFE_strncpy(key_string, "action", 99);
      trans_type=va_arg(ap, int);
      if (trans_type < 1000) {
        MC_SAFE_strncpy(value_string, MCVE_GetTypeString(trans_type), 1023);
      } else {
        MC_SAFE_strncpy(value_string, MCVE_GetEngineAdminTypeString(trans_type), 1023);
      }
      conn->queue[identifier].type=trans_type;
    break;

    case MC_ADMIN:
      MC_SAFE_strncpy(key_string, "admin", 99);
      trans_type=va_arg(ap, int);
      MC_SAFE_strncpy(value_string, MCVE_GetAdminTypeString(trans_type), 1023);
      conn->queue[identifier].admin=trans_type;
    break;

    case MC_USERNAME:
      MC_SAFE_strncpy(key_string, "username", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_PASSWORD:
      MC_SAFE_strncpy(key_string, "password", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_ACCOUNT:
      MC_SAFE_strncpy(key_string, "account", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_TRACKDATA:
      MC_SAFE_strncpy(key_string, "trackdata", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_EXPDATE:
      MC_SAFE_strncpy(key_string, "expdate", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_STREET:
      MC_SAFE_strncpy(key_string, "street", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_ZIP:
      MC_SAFE_strncpy(key_string, "zip", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_CV:
      MC_SAFE_strncpy(key_string, "cvv2", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_COMMENTS:
      MC_SAFE_strncpy(key_string, "comments", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_CLERKID:
      MC_SAFE_strncpy(key_string, "clerkid", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_STATIONID:
      MC_SAFE_strncpy(key_string, "stationid", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_APPRCODE:
      MC_SAFE_strncpy(key_string, "apprcode", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_AMOUNT:
      MC_SAFE_strncpy(key_string, "amount", 99);
      sprintf(value_string, "%.2f", va_arg(ap, double));
    break;

    case MC_PTRANNUM:
      MC_SAFE_strncpy(key_string, "ptrannum", 99);
      sprintf(value_string, "%ld", va_arg(ap, long));
    break;

    case MC_TTID:
      MC_SAFE_strncpy(key_string, "ttid", 99);
      sprintf(value_string, "%lld", va_arg(ap, MCVE_int64));
    break;

    case MC_USER:
      MC_SAFE_strncpy(key_string, "user", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_PWD:
      MC_SAFE_strncpy(key_string, "pwd", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_ACCT:
      MC_SAFE_strncpy(key_string, "acct", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_BDATE:
      MC_SAFE_strncpy(key_string, "bdate", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_EDATE:
      MC_SAFE_strncpy(key_string, "edate", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_BATCH:
      MC_SAFE_strncpy(key_string, "batch", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_FILE:
      MC_SAFE_strncpy(key_string, "file", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    case MC_AUDITTYPE:
      MC_SAFE_strncpy(key_string, "type", 99);
      trans_type=va_arg(ap, int);
      MC_SAFE_strncpy(value_string, MCVE_GetTypeString(trans_type), 1023);
    break;

    case MC_CUSTOM:
      MC_SAFE_strncpy(key_string, va_arg(ap, char *), 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;

    /* Additional args for restaurant, lodging and auto rental */
    case MC_EXAMOUNT:
      MC_SAFE_strncpy(key_string, "examount", 99);
      sprintf(value_string, "%.2f", va_arg(ap, double));
    break;
    
    case MC_EXCHARGES:
      MC_SAFE_strncpy(key_string, "excharges", 99);
      excharges_type = va_arg(ap, int);
      MC_SAFE_strncpy(value_string, MCVE_GetExchargesArgString(excharges_type), 1023);
    break;

    case MC_RATE:
      MC_SAFE_strncpy(key_string, "rate", 99);
      sprintf(value_string, "%.2f", va_arg(ap, double));
    break;

    case MC_RENTERNAME:
      MC_SAFE_strncpy(key_string, "rentername", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;    

    case MC_RETURNCITY:
      MC_SAFE_strncpy(key_string, "returncity", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;    

    case MC_RETURNSTATE:
      MC_SAFE_strncpy(key_string, "returnstate", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;    

    case MC_RETURNLOCATION:
      MC_SAFE_strncpy(key_string, "returnlocation", 99);
      MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
    break;   
    
    case MC_INQUIRY:
      MC_SAFE_strncpy(key_string, "inquiry", 99);
      temp_tf=va_arg(ap, int);
      if (temp_tf) {
        MC_SAFE_strncpy(value_string, "yes", 1023); 
      } else {
        MC_SAFE_strncpy(value_string, "no", 1023);
      }
    break;
    
    case MC_PRIORITY:
      MC_SAFE_strncpy(key_string, "priority", 99);
      MC_SAFE_strncpy(value_string, MCVE_GetPriorityString(va_arg(ap, int)), 1023);
    break;

    default:
      if (key >= 2000) {
        temp=MCVE_GetUserArgString(key);
	if (temp != NULL && strlen(temp)) {
          MC_SAFE_strncpy(key_string, temp, 99);
	  if (key == MC_USER_SUB) {
	    sprintf(value_string, "%d", va_arg(ap, int));
	  } else if (key == MC_USER_CARDTYPES) {
	    temp_str=MCVE_GetCardTypesString(va_arg(ap, int));
	    MC_SAFE_strncpy(value_string, temp_str, 1023);
	    free(temp_str);
	  } else if (key == MC_USER_MODE) {
	    temp_str=MCVE_GetModeString(va_arg(ap, int));
	    MC_SAFE_strncpy(value_string, temp_str, 1023);
	    free(temp_str);
	  } else {
            MC_SAFE_strncpy(value_string, va_arg(ap, char *), 1023);
	  }
	}
      }
    break;
  };
  va_end(ap);
  if (strlen(key_string) && strlen(value_string)) {
    /* Maintain compatability with MCVE 2.0-2.1 */
    if (strcasecmp(key_string, "ttid") == 0) {
      MCVE_TransParam_Add(myconn, identifier, "sid", value_string);
    }
    return(MCVE_TransParam_Add(myconn, identifier, key_string, value_string));
  }
  return(0);
}


long MCVE_Ping(MCVE_CONN *myconn)
{
//  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_PING);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

/* TRANSACTION ROUTINES */


long MCVE_Sale(MCVE_CONN *myconn, char *username, char *password, char *trackdata, char *account, char *expdate, double amount,
                               char *street, char *zip, char *cv, char *comments, char *clerkid, char *stationid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || (trackdata == NULL && (account == NULL || expdate == NULL)) || amount <= 0) {
    return(MCVE_ERROR);
  }
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_SALE);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_TRACKDATA, trackdata);
  MCVE_TransParam(myconn, identifier, MC_ACCOUNT, account);
  MCVE_TransParam(myconn, identifier, MC_EXPDATE, expdate);
  MCVE_TransParam(myconn, identifier, MC_STREET, street);
  MCVE_TransParam(myconn, identifier, MC_ZIP, zip);
  MCVE_TransParam(myconn, identifier, MC_CV, cv);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, amount);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_PreAuth(MCVE_CONN *myconn, char *username, char *password, char *trackdata, char *account, char *expdate, double amount,
                                     char *street, char *zip, char *cv, char *comments, char *clerkid, char *stationid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || (trackdata == NULL && (account == NULL || expdate == NULL)) || amount <= 0)
    return(MCVE_ERROR);


  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_PREAUTH);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_TRACKDATA, trackdata);
  MCVE_TransParam(myconn, identifier, MC_ACCOUNT, account);
  MCVE_TransParam(myconn, identifier, MC_EXPDATE, expdate);
  MCVE_TransParam(myconn, identifier, MC_STREET, street);
  MCVE_TransParam(myconn, identifier, MC_ZIP, zip);
  MCVE_TransParam(myconn, identifier, MC_CV, cv);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, amount);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_Void(MCVE_CONN *myconn, char *username, char *password, MCVE_int64 ttid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || (ttid == -1 && ptrannum == -1))
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_VOID);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);

  if (ttid != -1)
    MCVE_TransParam(myconn, identifier, MC_TTID, ttid);
  else if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_PreAuthCompletion(MCVE_CONN *myconn, char *username, char *password, double finalamount, MCVE_int64 ttid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || finalamount <= 0 || (ttid == -1 && ptrannum == -1))
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_PREAUTHCOMPLETE);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, finalamount);
  if (ttid != -1)
    MCVE_TransParam(myconn, identifier, MC_TTID, ttid);
  else if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Force(MCVE_CONN *myconn, char *username, char *password, char *trackdata, char *account, char *expdate, double amount,
                char *authcode, char *comments, char *clerkid, char *stationid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || (trackdata == NULL && (account == NULL || expdate == NULL)) || amount <= 0)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_FORCE);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_TRACKDATA, trackdata);
  MCVE_TransParam(myconn, identifier, MC_ACCOUNT, account);
  MCVE_TransParam(myconn, identifier, MC_EXPDATE, expdate);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, amount);
  MCVE_TransParam(myconn, identifier, MC_APPRCODE, authcode);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Return(MCVE_CONN *myconn, char *username, char *password, char *trackdata, char *account, char *expdate, double amount,
                                     char *comments, char *clerkid, char *stationid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (username == NULL || password == NULL || (trackdata == NULL && (account == NULL || expdate == NULL)) || amount <= 0)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_RETURN);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_TRACKDATA, trackdata);
  MCVE_TransParam(myconn, identifier, MC_ACCOUNT, account);
  MCVE_TransParam(myconn, identifier, MC_EXPDATE, expdate);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, amount);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_Override(MCVE_CONN *myconn, char *username, char *password, char *trackdata, char *account, char *expdate, double amount,
                                     char *street, char *zip, char *cv, char *comments, char *clerkid, char *stationid, long ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (username == NULL || password == NULL || (trackdata == NULL && (account == NULL || expdate == NULL)) || amount <= 0)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_OVERRIDE);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_TRACKDATA, trackdata);
  MCVE_TransParam(myconn, identifier, MC_ACCOUNT, account);
  MCVE_TransParam(myconn, identifier, MC_EXPDATE, expdate);
  MCVE_TransParam(myconn, identifier, MC_STREET, street);
  MCVE_TransParam(myconn, identifier, MC_ZIP, zip);
  MCVE_TransParam(myconn, identifier, MC_CV, cv);
  MCVE_TransParam(myconn, identifier, MC_AMOUNT, amount);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_Settle(MCVE_CONN *myconn, char *username, char *password, char *batch)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_SETTLE);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  if (batch == NULL || !strlen(batch)) {
    MCVE_TransParam(myconn, identifier, MC_BATCH, "all");
  } else {
    MCVE_TransParam(myconn, identifier, MC_BATCH, batch);
  }

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}



/* RESPONSE CHECKING ROUTINES */

int MCVE_ReturnStatus(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int status;
  char *code;
  code=MCVE_ResponseParam(myconn, identifier, "code");
  if (code == NULL) return(MCVE_SUCCESS);
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  if (conn->queue[identifier].code == -1)
    status=-1;
  else if (conn->queue[identifier].code == MCVE_SUCCESS)
    status=MCVE_SUCCESS;
  else if (conn->queue[identifier].code == MCVE_AUTH)
    status=MCVE_SUCCESS;
  else
    status=MCVE_FAIL;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(status);
}

int MCVE_ReturnCode(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int code;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    code=conn->queue[identifier].code;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(code);
}

long MCVE_TransactionItem(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long item=-1;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    if (conn->queue[identifier].item != NULL)
      item=atol(conn->queue[identifier].item);
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(item);
}

long MCVE_TransactionBatch(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long batch=-1;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    if (conn->queue[identifier].batch != NULL)
      batch=atol(conn->queue[identifier].batch);
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(batch);
}

MCVE_int64 MCVE_TransactionID(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  MCVE_int64 tid=-1;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    if (conn->queue[identifier].tid != -1)
      tid=conn->queue[identifier].tid;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(tid);
}

char *MCVE_TransactionAuth(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  char *auth=NULL;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    if (conn->queue[identifier].auth != NULL)
      auth=conn->queue[identifier].auth;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(auth);
}

char *MCVE_TransactionText(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  char *text=NULL;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    if (conn->queue[identifier].text != NULL)
      text=conn->queue[identifier].text;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(text);
}

int MCVE_TransactionAVS(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int code;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    code=conn->queue[identifier].avs;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(code);
}

int MCVE_TransactionCV(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int code;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
    code=conn->queue[identifier].cv;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(code);
}

long MCVE_TransInQueue(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long auths;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  auths=conn->outstanding_auths;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(auths);
}

int MCVE_CheckStatus(MCVE_CONN *myconn, long trans)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int status;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  status=conn->queue[trans].status;
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(status);
}

long MCVE_CompleteAuthorizations(MCVE_CONN *myconn, long **listings)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long num=0, i;
  MCVE_lock(myconn, MCVE_LOCK_QUEUE);
  for (i=0; i<conn->queue_length; i++) {
   if  (conn->queue[i].status == MCVE_DONE)
     num++;
  }
  if (num != 0) {
    listings[0] = (long *)malloc((num+1)*sizeof(long));
    num=0;
    for (i=0; i<conn->queue_length; i++) {
      if (conn->queue[i].status == MCVE_DONE) {
        listings[0][num]=i;
        num++;
      }
    }
  }
  MCVE_unlock(myconn, MCVE_LOCK_QUEUE);
  return(num);
}


/* USER-LEVEL ADMINISTRATIVE REQUESTS */

long MCVE_Gut(MCVE_CONN *myconn, char *username, char *password, int type, char *acct, char *clerkid, char *stationid, char *comments,
                              double ptrannum, char *bdate, char *edate)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_GUT);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_AUDITTYPE, type);
  MCVE_TransParam(myconn, identifier, MC_ACCT, acct);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_BDATE, bdate);
  MCVE_TransParam(myconn, identifier, MC_EDATE, edate);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_Gl(MCVE_CONN *myconn, char *username, char *password, int type, char *acct, char *batch, char *clerkid, char *stationid, char *comments,
                              double ptrannum, char *bdate, char *edate)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_GL);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_AUDITTYPE, type);
  MCVE_TransParam(myconn, identifier, MC_ACCT, acct);
  MCVE_TransParam(myconn, identifier, MC_BATCH, batch);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_BDATE, bdate);
  MCVE_TransParam(myconn, identifier, MC_EDATE, edate);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Gft(MCVE_CONN *myconn, char *username, char *password, int type, char *acct, char *clerkid, char *stationid, char *comments,
                              double ptrannum, char *bdate, char *edate)
{
  // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL)
    return(MCVE_ERROR);
    
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_GFT);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_AUDITTYPE, type);
  MCVE_TransParam(myconn, identifier, MC_ACCT, acct);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  MCVE_TransParam(myconn, identifier, MC_BDATE, bdate);
  MCVE_TransParam(myconn, identifier, MC_EDATE, edate);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

/*
long MCVE_Ub(MCVE_CONN *myconn, char *username, char *password)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  MCVE_TRANS trans;
  MCVE_ClearTransaction(&trans);
  if (username == NULL || password == NULL)
    return(MCVE_ERROR);
  trans.trans_type=MCVE_UB;
  trans.username=username;
  trans.password=password;
  return(MCVE_SendTransaction(myconn, trans));
}
*/

long MCVE_Rs(MCVE_CONN *myconn, char *username, char *password, char *stationid)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL || stationid == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_RS);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Qc(MCVE_CONN *myconn, char *username, char *password, char *clerkid, char *stationid, char *comments, double ptrannum)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_QC);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_CLERKID, clerkid);
  MCVE_TransParam(myconn, identifier, MC_STATIONID, stationid);
  MCVE_TransParam(myconn, identifier, MC_COMMENTS, comments);
  if (ptrannum != -1)
    MCVE_TransParam(myconn, identifier, MC_PTRANNUM, ptrannum);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);

}

long MCVE_Chkpwd(MCVE_CONN *myconn, char *username, char *password)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_CHKPWD);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Bt(MCVE_CONN *myconn, char *username, char *password)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (username == NULL || password == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADMIN);
  MCVE_TransParam(myconn, identifier, MC_ADMIN, MC_ADMIN_BT);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, username);
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

/* ENGINE-ADMIN REQUESTS */


long MCVE_Chngpwd(MCVE_CONN *myconn, char *adminpassword, char *newpassword)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (adminpassword == NULL || newpassword == NULL)
    return(MCVE_ERROR);

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_CHNGPWD);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpassword);
  MCVE_TransParam(myconn, identifier, MC_PWD, newpassword);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_ListUsers(MCVE_CONN *myconn, char *password)
{
  //  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (password == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_LISTUSERS);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_EnableUser(MCVE_CONN *myconn, char *adminpwd, char *user)
{
  // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (adminpwd == NULL || user == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ENABLEUSER);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  MCVE_TransParam(myconn, identifier, MC_USER, user);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_DisableUser(MCVE_CONN *myconn, char *adminpwd, char *user)
{
  // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (adminpwd == NULL || user == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_DISABLEUSER);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  MCVE_TransParam(myconn, identifier, MC_USER, user);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


long MCVE_GetUserInfo(MCVE_CONN *myconn, char *adminpwd, char *user)
{
  long identifier;
  if (adminpwd == NULL || user == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_GETUSERINFO);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  MCVE_TransParam(myconn, identifier, MC_USER, user);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_DelUser(MCVE_CONN *myconn, char *password, char *user)
{
  long identifier;
  if (password == NULL || user == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_DELUSER);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, password);
  MCVE_TransParam(myconn, identifier, MC_USER, user);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_ListStats(MCVE_CONN *myconn, char *adminpwd)
{
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (adminpwd == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_LISTSTATS);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Export(MCVE_CONN *myconn, char *adminpwd, char *filename)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (filename == NULL || adminpwd == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_EXPORT);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  MCVE_TransParam(myconn, identifier, MC_FILE, filename);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

long MCVE_Import(MCVE_CONN *myconn, char *adminpwd, char *filename)
{
 // _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;
  if (filename == NULL || adminpwd == NULL)
    return(MCVE_ERROR);
  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_IMPORT);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);
  MCVE_TransParam(myconn, identifier, MC_FILE, filename);
  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

char *MCVE_GetUserParam(MCVE_CONN *myconn, long identifier, int key)
{
  char *str_key=NULL;
  str_key=MCVE_GetUserArgString(key);
  if (str_key == NULL) return(NULL);
  return(MCVE_ResponseParam(myconn, identifier, str_key));
}


/* COMMA DELIMITED ROUTINES */

int MCVE_IsCommaDelimited(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].iscommadelimiated);
}


int MCVE_ParseCommaDelimited(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int columns;
  long rows;
  int i;
  long j;
  int offset;
  char *temp, *temp2, *str;
  char *line_temp1, *line_temp2;
  columns=MCVE_Count_CD_Columns(myconn, identifier);
  rows=MCVE_Count_CD_Lines(myconn, identifier);
  conn->queue[identifier].separated=(char ***)malloc(rows*sizeof(char **));
  for (j=0; j<rows; j++) {
    conn->queue[identifier].separated[j]=(char **)malloc(columns*sizeof(char *));
  }
  temp=conn->queue[identifier].response;
  for (j=0; j<rows; j++) {
    str=MCVE_Get_CD_Line(temp);
    if (str == NULL) break;
    offset=strlen(str)+2;
    temp2=temp+offset;
    temp=temp2;
    line_temp1=str;
    for (i=0; i<columns; i++) {
      line_temp2=strchr(line_temp1, ',');
      if (line_temp2 == NULL) {
        conn->queue[identifier].separated[j][i]=MCVE_midstr(line_temp1, 0, strlen(line_temp1));
        break;
      }
      conn->queue[identifier].separated[j][i]=MCVE_midstr(line_temp1, 0, strlen(line_temp1)-strlen(line_temp2));
      line_temp1=line_temp2+1;
    }
    free(str);
  }
  conn->queue[identifier].columns=columns;
  conn->queue[identifier].rows=rows-1;
  return(1);
}

char *MCVE_GetCommaDelimited(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].response);
}


char *MCVE_GetCell(MCVE_CONN *myconn, long identifier, char *column, long row)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int i;
  for (i=0; i<conn->queue[identifier].columns; i++) {
    if (strcasecmp(column, conn->queue[identifier].separated[0][i]) == 0)
      return(conn->queue[identifier].separated[row+1][i]);
  }
  return(NULL);
}

char *MCVE_GetCellByNum(MCVE_CONN *myconn, long identifier, int column, long row)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].separated[row+1][column]);
}

int MCVE_NumColumns(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].columns);
}

long MCVE_NumRows(MCVE_CONN *myconn, long identifier)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].rows);
}

char *MCVE_GetHeader(MCVE_CONN *myconn, long identifier, int column_num)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  return(conn->queue[identifier].separated[0][column_num]);
}


/* LEGACY USER FUNCTIONS */

typedef struct _UserSetup {
  char *proc;
  char *user;
  char *pwd;
  char *indcode;
  char *merchid;
  char *bankid;
  char *termid;
  char *clientnum;
  char *storeid;
  char *agentid;
  char *chainid;
  char *zipcode;
  char *timezone;
  char *merchcat;
  char *mername;
  char *merchloc;
  char *statecode;
  char *servicephone;
} _UserSetup;

void MCVE_InitUserSetup(MCVE_UserSetup *setup)
{

  _UserSetup *usersetup;
  setup[0]=(MCVE_UserSetup)malloc(sizeof(_UserSetup)*1);
  usersetup=(_UserSetup *)setup[0];
  usersetup->proc=NULL;
  usersetup->user=NULL;
  usersetup->pwd=NULL;
  usersetup->indcode=NULL;
  usersetup->merchid=NULL;
  usersetup->bankid=NULL;
  usersetup->termid =NULL;
  usersetup->clientnum =NULL;
  usersetup->storeid =NULL;
  usersetup->agentid =NULL;
  usersetup->chainid =NULL;
  usersetup->zipcode =NULL;
  usersetup->timezone =NULL;
  usersetup->merchcat =NULL;
  usersetup->mername =NULL;
  usersetup->merchloc =NULL;
  usersetup->statecode =NULL;
  usersetup->servicephone =NULL;
}

int MCVE_AddUserArg(MCVE_UserSetup *setup, int argtype, char *argval)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];

  char **field;

  if (argtype == MCVE_PROC)
    field=&usersetup->proc;
  else if (argtype == MCVE_USER)
    field=&usersetup->user;
  else if (argtype == MCVE_PWD)
    field=&usersetup->pwd;
  else if (argtype == MCVE_INDCODE)
    field=&usersetup->indcode;
  else if (argtype == MCVE_MERCHID)
    field=&usersetup->merchid;
  else if (argtype == MCVE_BANKID)
    field=&usersetup->bankid;
  else if (argtype == MCVE_TERMID)
    field=&usersetup->termid;
  else if (argtype == MCVE_CLIENTNUM)
    field=&usersetup->clientnum;
  else if (argtype == MCVE_STOREID)
    field=&usersetup->storeid;
  else if (argtype == MCVE_AGENTID)
    field=&usersetup->agentid;
  else if (argtype == MCVE_CHAINID)
    field=&usersetup->chainid;
  else if (argtype == MCVE_ZIPCODE)
    field=&usersetup->zipcode;
  else if (argtype == MCVE_TIMEZONE)
    field=&usersetup->timezone;
  else if (argtype == MCVE_MERCHCAT)
    field=&usersetup->merchcat;
  else if (argtype == MCVE_MERNAME)
    field=&usersetup->mername;
  else if (argtype == MCVE_MERCHLOC)
    field=&usersetup->merchloc;
  else if (argtype == MCVE_STATECODE)
    field=&usersetup->statecode;
  else if (argtype == MCVE_SERVICEPHONE)
    field=&usersetup->servicephone;
  else
    return(0);

  field[0]=(char *)malloc((strlen(argval)+1)*sizeof(char));
  strcpy(field[0], argval);
  return(1);
}

/*
char *MCVE_GetUserArg(MCVE_UserSetup *setup, int argtype)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];

  if (argtype == MCVE_PROC)
    return(usersetup->proc);
  else if (argtype == MCVE_USER)
    return(usersetup->user);
  else if (argtype == MCVE_PWD)
    return(usersetup->pwd);
  else if (argtype == MCVE_INDCODE)
    return(usersetup->indcode);
  else if (argtype == MCVE_MERCHID)
    return(usersetup->merchid);
  else if (argtype == MCVE_BANKID)
    return(usersetup->bankid);
  else if (argtype == MCVE_TERMID)
    return(usersetup->termid);
  else if (argtype == MCVE_CLIENTNUM)
    return(usersetup->clientnum);
  else if (argtype == MCVE_STOREID)
    return(usersetup->storeid);
  else if (argtype == MCVE_AGENTID)
    return(usersetup->agentid);
  else if (argtype == MCVE_CHAINID)
    return(usersetup->chainid);
  else if (argtype == MCVE_ZIPCODE)
    return(usersetup->zipcode);
  else if (argtype == MCVE_TIMEZONE)
    return(usersetup->timezone);
  else if (argtype == MCVE_MERCHCAT)
    return(usersetup->merchcat);
  else if (argtype == MCVE_MERNAME)
    return(usersetup->mername);
  else if (argtype == MCVE_MERCHLOC)
    return(usersetup->merchloc);
  else if (argtype == MCVE_STATECODE)
    return(usersetup->statecode);
  else if (argtype == MCVE_SERVICEPHONE)
    return(usersetup->servicephone);

  return(NULL);
}
*/

void MCVE_DeleteUserSetup(MCVE_UserSetup *setup)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];

  if (usersetup->proc != NULL) {
    free(usersetup->proc);
    usersetup->proc=NULL;
  }
  if (usersetup->user != NULL) {
    free(usersetup->user);
    usersetup->user=NULL;
  }
  if (usersetup->pwd != NULL) {
    free(usersetup->pwd);
    usersetup->pwd=NULL;
  }
  if (usersetup->indcode != NULL) {
    free(usersetup->indcode);
    usersetup->indcode=NULL;
  }
  if (usersetup->merchid != NULL) {
    free(usersetup->merchid);
    usersetup->merchid=NULL;
  }
  if (usersetup->bankid != NULL) {
    free(usersetup->bankid);
    usersetup->bankid=NULL;
  }
  if (usersetup->termid != NULL) {
    free(usersetup->termid);
    usersetup->termid=NULL;
  }
  if (usersetup->clientnum != NULL) {
    free(usersetup->clientnum);
    usersetup->clientnum=NULL;
  }
  if (usersetup->storeid != NULL) {
    free(usersetup->storeid);
    usersetup->storeid=NULL;
  }
  if (usersetup->agentid != NULL) {
    free(usersetup->agentid);
    usersetup->agentid=NULL;
  }
  if (usersetup->chainid != NULL) {
    free(usersetup->chainid);
    usersetup->chainid=NULL;
  }
  if (usersetup->zipcode != NULL) {
    free(usersetup->zipcode);
    usersetup->zipcode=NULL;
  }
  if (usersetup->timezone != NULL) {
    free(usersetup->timezone);
    usersetup->timezone=NULL;
  }
  if (usersetup->merchcat != NULL) {
    free(usersetup->merchcat);
    usersetup->merchcat=NULL;
  }
  if (usersetup->mername != NULL) {
    free(usersetup->mername);
    usersetup->mername=NULL;
  }
  if (usersetup->merchloc != NULL) {
    free(usersetup->merchloc);
    usersetup->merchloc=NULL;
  }
  if (usersetup->statecode != NULL) {
    free(usersetup->statecode);
    usersetup->statecode=NULL;
  }
  if (usersetup->servicephone != NULL) {
    free(usersetup->servicephone);
    usersetup->servicephone=NULL;
  }
  free(setup[0]);
}

long MCVE_AddUser(MCVE_CONN *myconn, char *adminpwd, MCVE_UserSetup *setup)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_ADDUSER);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);

  if (usersetup->proc != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PROC, usersetup->proc);
  }
  if (usersetup->user != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_USER, usersetup->user);
  }
  if (usersetup->pwd != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PWD, usersetup->pwd);
  }
  if (usersetup->indcode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_INDCODE, usersetup->indcode);
  }
  if (usersetup->merchid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHID, usersetup->merchid);
  }
  if (usersetup->bankid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_BANKID, usersetup->bankid);
  }
  if (usersetup->termid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_TERMID, usersetup->termid);
  }
  if (usersetup->clientnum != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_CLIENTNUM, usersetup->clientnum);
  }
  if (usersetup->storeid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_STOREID, usersetup->storeid);
  }
  if (usersetup->agentid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_AGENTID, usersetup->agentid);
  }
  if (usersetup->chainid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_CHAINID, usersetup->chainid);
  }
  if (usersetup->zipcode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_ZIPCODE, usersetup->zipcode);
  }
  if (usersetup->timezone != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_TIMEZONE, usersetup->timezone);
  }
  if (usersetup->merchcat != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHCAT, usersetup->merchcat);
  }
  if (usersetup->mername != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERNAME, usersetup->mername);
  }
  if (usersetup->merchloc != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHLOC, usersetup->merchloc);
  }
  if (usersetup->statecode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_STATECODE, usersetup->statecode);
  }
  if (usersetup->servicephone != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PHONE, usersetup->servicephone);
  }

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}

char *MCVE_GetUserArg(MCVE_UserSetup *setup, int argtype)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];
  char *field;

  if (argtype == MCVE_PROC)
    field=usersetup->proc;
  else if (argtype == MCVE_USER)
    field=usersetup->user;
  else if (argtype == MCVE_PWD)
    field=usersetup->pwd;
  else if (argtype == MCVE_INDCODE)
    field=usersetup->indcode;
  else if (argtype == MCVE_MERCHID)
    field=usersetup->merchid;
  else if (argtype == MCVE_BANKID)
    field=usersetup->bankid;
  else if (argtype == MCVE_TERMID)
    field=usersetup->termid;
  else if (argtype == MCVE_CLIENTNUM)
    field=usersetup->clientnum;
  else if (argtype == MCVE_STOREID)
    field=usersetup->storeid;
  else if (argtype == MCVE_AGENTID)
    field=usersetup->agentid;
  else if (argtype == MCVE_CHAINID)
    field=usersetup->chainid;
  else if (argtype == MCVE_ZIPCODE)
    field=usersetup->zipcode;
  else if (argtype == MCVE_TIMEZONE)
    field=usersetup->timezone;
  else if (argtype == MCVE_MERCHCAT)
    field=usersetup->merchcat;
  else if (argtype == MCVE_MERNAME)
    field=usersetup->mername;
  else if (argtype == MCVE_MERCHLOC)
    field=usersetup->merchloc;
  else if (argtype == MCVE_STATECODE)
    field=usersetup->statecode;
  else if (argtype == MCVE_SERVICEPHONE)
    field=usersetup->servicephone;
  else
    return(NULL);
  return(field);
}

long MCVE_EditUser(MCVE_CONN *myconn, char *adminpwd, MCVE_UserSetup *setup)
{
  _UserSetup *usersetup=(_UserSetup *)setup[0];
  //_MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier;

  identifier=MCVE_TransNew(myconn);
  MCVE_TransParam(myconn, identifier, MC_TRANTYPE, MC_TRAN_EDITUSER);
  MCVE_TransParam(myconn, identifier, MC_USERNAME, "MCVEADMIN");
  MCVE_TransParam(myconn, identifier, MC_PASSWORD, adminpwd);

  if (usersetup->proc != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PROC, usersetup->proc);
  }
  if (usersetup->user != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_USER, usersetup->user);
  }
  if (usersetup->pwd != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PWD, usersetup->pwd);
  }
  if (usersetup->indcode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_INDCODE, usersetup->indcode);
  }
  if (usersetup->merchid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHID, usersetup->merchid);
  }
  if (usersetup->bankid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_BANKID, usersetup->bankid);
  }
  if (usersetup->termid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_TERMID, usersetup->termid);
  }
  if (usersetup->clientnum != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_CLIENTNUM, usersetup->clientnum);
  }
  if (usersetup->storeid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_STOREID, usersetup->storeid);
  }
  if (usersetup->agentid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_AGENTID, usersetup->agentid);
  }
  if (usersetup->chainid != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_CHAINID, usersetup->chainid);
  }
  if (usersetup->zipcode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_ZIPCODE, usersetup->zipcode);
  }
  if (usersetup->timezone != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_TIMEZONE, usersetup->timezone);
  }
  if (usersetup->merchcat != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHCAT, usersetup->merchcat);
  }
  if (usersetup->mername != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERNAME, usersetup->mername);
  }
  if (usersetup->merchloc != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_MERCHLOC, usersetup->merchloc);
  }
  if (usersetup->statecode != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_STATECODE, usersetup->statecode);
  }
  if (usersetup->servicephone != NULL) {
    MCVE_TransParam(myconn, identifier, MC_USER_PHONE, usersetup->servicephone);
  }

  if (!MCVE_TransSend(myconn, identifier))
    return(-1);
  return(identifier);
}


/* TEXT RESPONSE CONVERSIONS */

char *MCVE_TEXT_Code(int code)
{
  char *text=NULL;
  switch (code) {
    case MCVE_AUTH:
      text="AUTH";
    break;
    case MCVE_DENY:
      text="DENY";
    break;
    case MCVE_DUPL:
      text="DUPLICATE TRANS";
    break;
    case MCVE_CALL:
      text="CALL PROCESSOR";
    break;
    case MCVE_PKUP:
      text="PICKUP CARD";
    break;
    case MCVE_RETRY:
      text="RETRY";
    break;
    case MCVE_SETUP:
      text="SETUP";
    break;
    case MCVE_TIMEOUT:
      text="TIMEOUT";
    break;
    case MCVE_SUCCESS:
      text="SUCCESS";
    break;
    case MCVE_FAIL:
      text="FAIL";
    break;
    case MCVE_ERROR:
      text="ERROR";
    break;
    default:
      text="UNKNOWN";
    break;
  }
  return(text);
}

char *MCVE_TEXT_AVS(int code)
{
  char *text=NULL;
  switch(code) {
    case MCVE_STREET:
      text="STREET FAILED";
    break;
    case MCVE_ZIP:
      text="ZIP FAILED";
    break;
    case MCVE_GOOD:
      text="GOOD";
    break;
    case MCVE_BAD:
      text="BAD";
    break;
    case MCVE_UNKNOWN:
      text="UNKNOWN";
    break;
    default:
      text="UNKNOWN";
    break;
  } 
  return(text);
}

char *MCVE_TEXT_CV(int code)
{
 return(MCVE_TEXT_AVS(code));
}








