/* MCVE v3.x C API
   (c) 2002 by Main Street Softworks, Inc.
   Written by Brad House

   This API is being released to the public domain to be modified and used in any manor the holder of this code sees fit.
   For any questions please contact  support@mainstreetsoftworks.com
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "libmcve_main.h"
#include "mcve.h"


// Standard includes ... all platforms
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>

//Win32 includes
#ifdef WIN32
#include <io.h>
#include <winsock2.h>
#include <sys/timeb.h>
#else // !Win32
// UNIX/POSIX includes
#include <unistd.h>
#include <sys/socket.h>
#ifndef DISABLE_IP
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <sys/time.h>
#include <dirent.h>
#ifdef MCVE_PTHREAD
#include <pthread.h>
#endif // MCVE_PTHREAD

#endif //Win32

//SSL includes
#ifdef ENABLE_SSL
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#endif
#include "libmcve_sockets.h"
#include "libmcve_misc.h"
#include "libmcve_locking.h"

#ifdef WIN32
#define ioctl(a,b,c) ioctlsocket(a,b,c)
#endif

#ifndef DISABLE_IP

int MCVE_InitSockets()
{
#ifdef WIN32
  WSADATA wsaData;

  if (WSAStartup(0x202,&wsaData) == SOCKET_ERROR) {
    fprintf(stderr,"WSAStartup failed with error %d\n",WSAGetLastError());
    WSACleanup();
    return 0;
  }
#endif // WIN32
  return(1);
}

void MCVE_CloseSocket(int fd)
{
#ifdef WIN32
  closesocket(fd);
#else
  close(fd);
#endif
}

int MCVE_SetNonBlock(int fd, int tf)
{
#ifdef O_NONBLOCK
  int flags;

  if ((flags=fcntl(fd, F_GETFL, 0)) == -1)
    return(0);
  if (tf) flags |= O_NONBLOCK;
  else flags &= ~(O_NONBLOCK);
  if (fcntl(fd, F_SETFL, flags) == -1)
    return(0);
#else
  if (ioctl(fd, FIONBIO, &tf) == -1)
    return(0);
#endif
  return(1);
}



int MCVE_Real_Connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen, MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  int rc;
  int arglen;
  fd_set fds;
  struct timeval timeout;

  if (!conn->max_conn_time) { // Blocking mode, let OS decide when to stop
    if (connect(sockfd, serv_addr, addrlen) == -1) {
      MCVE_Set_Conn_Error(myconn, strerror(errno));  // Set Error Text
      return(0);
    } else {
      return(1);
    }
  } else { // Non-blocking mode, stop after maxtime (seconds)
    if (!MCVE_SetNonBlock(sockfd, 1)) { // Set non-block mode
      MCVE_Set_Conn_Error(myconn, "Could not set Non-Blocking mode");  // Set Error Text
      return(0);
    }
    rc=connect(sockfd, serv_addr, addrlen);
    if (rc >= 0) {   //  connect() completed immediately
      MCVE_SetNonBlock(sockfd, 0);  // Set blocking mode
      return(1);
#ifdef WIN32
    } else if (WSAGetLastError() != WSAEWOULDBLOCK) {
      MCVE_Set_Conn_Error(myconn, "WinSock32 Unknown Error");
      return(0);
#else
    } else if (errno != EINPROGRESS) { // Error occurred
      MCVE_Set_Conn_Error(myconn, strerror(errno));  // Set Error Text
      return(0);
#endif
    } else {  // Connect in progress
      FD_ZERO(&fds);
      FD_SET(sockfd, &fds);
      timeout.tv_sec=conn->max_conn_time;
      timeout.tv_usec=0;
      rc=select(sockfd+1, NULL, &fds, NULL, &timeout);
      if (rc == 0) { // Connection taking too long
        MCVE_Set_Conn_Error(myconn, "Connection Timed Out");  // Set Error Text
        return(0);
      } else if (rc < 0) {  // Error Occurred
        MCVE_Set_Conn_Error(myconn, "Unknown error occurred"); // Set Error Text
        return(0);
      } else {  // rc > 0 .. good so far...
        rc=0;
        arglen=sizeof(int);
        if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &rc, &arglen) < 0)
          rc=errno;
        if (rc) {
          MCVE_Set_Conn_Error(myconn, strerror(rc));  // Set Error Text
          return(0);  // Connect Failed
        }
      }
    }
    MCVE_SetNonBlock(sockfd, 0);  // Set blocking mode
    return(1);
  }
  return(0); // Should never get here...
}

long MCVE_read(int fd, char *buf, long bytes)
{
#ifdef WIN32
  return(recv(fd, buf, bytes, 0));
#else
  return(read(fd, buf, bytes));
#endif
}

long MCVE_write(int fd, char *buf, long bytes)
{
#ifdef WIN32
  return(send(fd, buf, bytes, 0));
#else
  return(write(fd, buf, bytes));
#endif
}

int MCVE_ip_connect(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  struct sockaddr_in peer;
  struct hostent *hp;
  int soct;
  int ghbn_failed=0;

  hp = gethostbyname(conn->location);
  if (hp==NULL)  ghbn_failed=1;  // Fails on Darwin (MacOSX) for local IP addresses....gotta handle it
  memset(&peer, 0, sizeof(peer));
  peer.sin_family = AF_INET;
  peer.sin_port = htons(conn->port);
  if (!ghbn_failed)
    memcpy((char *)&peer.sin_addr, hp->h_addr_list[0], hp->h_length);
  else {
    peer.sin_addr.s_addr = inet_addr(conn->location);
    if (peer.sin_addr.s_addr == -1)
      return(-1);
  }
  soct = socket(AF_INET, SOCK_STREAM, 0);
  MCVE_Set_Conn_Error(myconn, strerror(errno));  // Set Error Text
  if (soct == -1) {
    return(-1);
  }
  if (!MCVE_Real_Connect(soct, (struct sockaddr *)&peer, sizeof(peer), myconn)) {
    MCVE_CloseSocket(soct);
    return(-1);
  }
  return(soct);
}

int MCVE_VerifyPing(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  long identifier=-1;
  time_t t, lastt;
  int status=0;
  int blocking=0;

  blocking=conn->blocking;

  MCVE_SetBlocking(myconn, 0);

  identifier=MCVE_Ping(myconn);
  time(&lastt);
  t=lastt;
  status=1;
  while (MCVE_CheckStatus(myconn, identifier) != MCVE_DONE) {
     if (!MCVE_Monitor(myconn)) {
       status=0; break;
     }
     time(&t);
     if ((t-lastt) > 3) { // Response Timeout
       status=0; break;
      }
      MCVE_uwait(10000);
   }
   
   MCVE_SetBlocking(myconn, blocking);

   MCVE_DeleteResponse(myconn, identifier);
   if (!status) {

      return(0);
   }
   return(1);
}

// Let's see if there's anything in the buffer to read!
int MCVE_CheckRead(int ptr, long delay)
{
  fd_set readfs;
  struct timeval timeout;
  int res;

  FD_ZERO(&readfs);
  FD_SET(ptr, &readfs);
  timeout.tv_sec=0;
  timeout.tv_usec=delay;
  res=select(ptr+1, &readfs, NULL, NULL, &timeout);
  if (res > 0) {
       if (FD_ISSET(ptr, &readfs)) {
         return(1);
       }
  }
  return(0);
}


// Let's make sure that the buffer isn't too full to write to
//  I'd assume this will only happen in dial-up scenerios
int MCVE_CheckWrite(int ptr, long delay)
{
  fd_set sendfs;
  struct timeval timeout;
  int res;

  FD_ZERO(&sendfs);
  FD_SET(ptr, &sendfs);
  timeout.tv_sec=0;
  timeout.tv_usec=delay;

  res=select(ptr+1, NULL, &sendfs, NULL, &timeout);
  if (res > 0) {
       if (FD_ISSET(ptr, &sendfs)) {
         return(1);
       }
  }
  return(0);
}

int MCVE_Monitor_IP(MCVE_CONN *myconn)
{
  _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
  char temp_buffer[1025];
  int bytes_read=0;
  int status=1;
#ifdef ENABLE_SSL
  int err;
#endif

  // Let's read the whole buffer, 1K at a time
  while (MCVE_CheckRead(conn->ptr, 10)
#ifdef ENABLE_SSL
    || ((conn->conn_method == MCVE_SSL) && (SSL_pending((SSL *)conn->ssl)))
#endif
    ) {
     memset(temp_buffer, 0, 1025);
     if (conn->conn_method == MCVE_SOCKETS) {
        bytes_read=MCVE_read(conn->ptr, temp_buffer, 1024);
     } else if (conn->conn_method == MCVE_SSL) {
#ifdef ENABLE_SSL
        bytes_read=SSL_read((SSL *)conn->ssl, temp_buffer, 1024);
    //    printf("Read: %s\r\n", temp_buffer);
#endif
     }
     if (bytes_read <= 0) {
       // connection closed
       status=0;
       break;
     }
     MCVE_lock(myconn, MCVE_LOCK_IN);
     if (conn->inbuf == NULL) {
       conn->inbuf=(char *)malloc(bytes_read+1);
       strcpy(conn->inbuf, temp_buffer);
     } else {
       conn->inbuf=(char *)realloc(conn->inbuf, strlen(conn->inbuf)+bytes_read+1);
       strcat(conn->inbuf, temp_buffer);
     }

     MCVE_unlock(myconn, MCVE_LOCK_IN);
  }

  MCVE_lock(myconn, MCVE_LOCK_OUT);
  if (status && conn->outbuf != NULL) {
  //  printf("About to write buffer of : %s\r\n", conn->outbuf);
    // We definately don't want to overflow the buffer.
    while (conn->outbuf != NULL) {
      if (MCVE_CheckWrite(conn->ptr, 10)) {
         if (conn->conn_method == MCVE_SOCKETS) {
            //write(conn->ptr, &*conn->outbuf, 1);
            MCVE_write(conn->ptr, conn->outbuf, strlen(conn->outbuf));
         } else if (conn->conn_method == MCVE_SSL) {
#ifdef ENABLE_SSL
           //  err=SSL_write((SSL *)conn->ssl, &*conn->outbuf, 1);
             err=SSL_write((SSL *)conn->ssl, conn->outbuf, strlen(conn->outbuf));
         //    printf("Wrote: %c -- %d\r\n", *conn->outbuf, err);
#endif
         }
         //conn->outbuf+=1;
         free(conn->outbuf);
         conn->outbuf=NULL;
      }
    }
 /*   if (*conn->outbuf == 0) {
      conn->outbuf=0;
      free(conn->outbuf);
      conn->outbuf=NULL;
    }  */
  }
  MCVE_unlock(myconn, MCVE_LOCK_OUT);
  return(status);
}

int MCVE_SendTransaction_IP(MCVE_CONN *myconn, char *identifier, char *transaction)
{
   _MCVE_CONN *conn=(_MCVE_CONN *)myconn[0];
   int size;
   size=strlen(identifier)+strlen(transaction)+4;
   MCVE_lock(myconn, MCVE_LOCK_OUT);
      if (conn->outbuf == NULL) {
        conn->outbuf=(char *)malloc(size);
        conn->outbuf[0]=0;
      } else
        conn->outbuf=(char *)realloc(conn->outbuf, strlen(conn->outbuf)+size);

      strcat(conn->outbuf, "\x02");  // STX  (start of transmission block)
      strcat(conn->outbuf, identifier);
      strcat(conn->outbuf, "\x1c");  // FS (field seperator in transmission block)
      strcat(conn->outbuf, transaction);
      strcat(conn->outbuf, "\x03");  // ETX (end of transmission block)
   MCVE_unlock(myconn, MCVE_LOCK_OUT);
   return(1);
}

#endif  /* DISABLE_IP */
